// ai.vala
//
// Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
//
//  This program is free software: you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see
//  <http://www.gnu.org/licenses/>.

namespace Tarot {

    public abstract class AiGame: Object {
        private Tarot.Game m_game;

        public signal void event_bid (uint bid);
        public signal void event_decl (bool decl);
        public signal void event_call (uint card);
        public signal void event_discard (uint[] cards);
        public signal void event_handful (uint[] cards);
        public signal void event_play (uint card);

        public signal void finished (int my_score);

        protected abstract Tarot.GameEvent get_best (Tarot.Game game);

        private static Tarot.Card[] to_cards (uint[] cards) {
            var ret = new Tarot.Card[cards.length];
            for (size_t i = 0; i < cards.length; i++) {
                ret[i] = (Tarot.Card) cards[i];
            }
            return ret;
        }

        private void add_event_or_fail (Tarot.GameEvent e) {
            if (add_event (e) != Tarot.GameError.OK) {
                stderr.printf ("Invalid server event: %s.\n", e.type ().to_string ());
                GLib.Process.abort ();
            }
        }

        private bool subscribed = false;

        public AiGame (Seat seat) {
            m_game = new Tarot.Game.alloc ();
            seat.event_setup.connect ((n_players, with_call) => {
                    var e = new Tarot.GameEvent.alloc_setup ((size_t) n_players, with_call);
                    add_event_or_fail (e);
                });
            seat.event_deal.connect ((icards) => {
                    Tarot.Player me = 0;
                    try {
                        me = (Tarot.Player) seat.me ();
                    } catch (GLib.Error e) {
                        stderr.printf (_ ("Could not get my player number: %s\n"), e.message);
                    }
                    var e = new Tarot.GameEvent.alloc_deal (me, to_cards (icards));
                    add_event_or_fail (e);
                });
            seat.event_bid.connect ((bid) => {
                    var e = new Tarot.GameEvent.alloc_bid ((Tarot.Bid) bid);
                    add_event_or_fail (e);
                });
            seat.event_decl.connect ((decl) => {
                    var e = new Tarot.GameEvent.alloc_decl (decl);
                    add_event_or_fail (e);
                });
            seat.event_call.connect ((call) => {
                    var e = new Tarot.GameEvent.alloc_call (call);
                    add_event_or_fail (e);
                });
            seat.event_dog.connect ((icards) => {
                    var e = new Tarot.GameEvent.alloc_dog (to_cards (icards));
                    add_event_or_fail (e);
                });
            seat.event_discard.connect ((icards) => {
                    var e = new Tarot.GameEvent.alloc_discard (to_cards (icards));
                    add_event_or_fail (e);
                });
            seat.event_handful.connect ((icards) => {
                    var e = new Tarot.GameEvent.alloc_handful (to_cards (icards));
                    add_event_or_fail (e);
                });
            seat.event_play.connect ((card) => {
                    var e = new Tarot.GameEvent.alloc_card (card);
                    add_event_or_fail (e);
                });
            seat.next.connect (() => {
                    next ();
                });
            seat.end.connect (scores => {
                    uint me = 0;
                    try {
                        me = seat.me ();
                    } catch (GLib.Error e) {
                        stderr.printf (_ ("Could not get my player number: %s\n"), e.message);
                    }
                    finished (scores[me]);
                });
            event_bid.connect (bid => {
                    try {
                        seat.bid (bid);
                    } catch (GLib.Error e) {
                        stderr.printf (_ ("Could not play: %s\n"), e.message);
                    }
                });
            event_decl.connect (decl => {
                    try {
                        seat.declare (decl);
                    } catch (GLib.Error e) {
                        stderr.printf (_ ("Could not play: %s\n"), e.message);
                    }
                });
            event_call.connect (card => {
                    try {
                        seat.call (card);
                    } catch (GLib.Error e) {
                        stderr.printf (_ ("Could not play: %s\n"), e.message);
                    }
                });
            event_discard.connect (cards => {
                    try {
                        seat.discard (cards);
                    } catch (GLib.Error e) {
                        stderr.printf (_ ("Could not play: %s\n"), e.message);
                    }
                });
            event_handful.connect (cards => {
                    try {
                        seat.handful (cards);
                    } catch (GLib.Error e) {
                        stderr.printf (_ ("Could not play: %s\n"), e.message);
                    }
                });
            event_play.connect (card => {
                    try {
                        seat.play (card);
                    } catch (GLib.Error e) {
                        stderr.printf (_ ("Could not play: %s\n"), e.message);
                    }
                });
            try {
                subscribed = true;
                seat.subscribe ();
            } catch (GLib.Error e) {
                stderr.printf (_ ("Could not subscribe to the signals: %s\n"), e.message);
            }
        }

        public void next () {
            assert (subscribed);
            Tarot.Player myself;
            if (m_game.get_main_player (out myself) != Tarot.GameError.OK) {
                // Not received the deal yet
                assert (m_game.step () == Tarot.Step.SETUP);
            } else {
                Tarot.Player next;
                if (m_game.get_next (out next) == Tarot.GameError.OK) {
                    if (myself == next) {
                        // I have to play.
                        Tarot.GameEvent best = get_best (m_game);
                        Tarot.Bid b;
                        bool d;
                        Tarot.Card c;
                        Tarot.Card[] cs;
                        if (best.get_bid (out b) == Tarot.GameEventError.OK) {
                            event_bid (b);
                        }
                        if (best.get_decl (out d) == Tarot.GameEventError.OK) {
                            event_decl (d);
                        }
                        if (best.get_call (out c) == Tarot.GameEventError.OK) {
                            event_call (c);
                        }
                        if (best.get_discard_alloc (out cs) == Tarot.GameEventError.OK) {
                            event_discard ((uint[]) cs);
                        }
                        if (best.get_handful_alloc (out cs) == Tarot.GameEventError.OK) {
                            event_handful ((uint[]) cs);
                        }
                        if (best.get_card (out c) == Tarot.GameEventError.OK) {
                            event_play (c);
                        }
                    }
                }
            }
        }

        private Tarot.GameError add_event (Tarot.GameEvent e) {
            size_t n;
            if (e.get_deal_all (out n, 0, {}) == Tarot.GameEventError.OK) {
                // The MCTS cannot react to a server-side game
                return Tarot.GameError.INVEV;
            }
            Tarot.GameError err = m_game.add_event (e);
            return err;
        }
    }

    public class StackingGame: AiGame {
        private Tarot.Predictor m_predictor;

        public StackingGame (Seat seat) {
            base (seat);
            m_predictor = new Tarot.Predictor.alloc ();
        }

        protected override Tarot.GameEvent get_best (Tarot.Game game) {
            m_predictor.set_game (game);
            double score;
            return m_predictor.best (out score);
        }
    }
}
